#include "expr.hpp"
#include "value.hpp"
#include "env.hpp"
#include "parse.hpp"
#include "catch.hpp"

NumExpr::NumExpr(int rep) {
  this->rep = rep;
  val = NEW(NumVal)(rep);
}

bool NumExpr::equals(PTR(Expr) other_expr) {
  PTR(NumExpr) n = CAST(NumExpr)(other_expr);
  if (n == NULL)
    return false;
  else
    return rep == n->rep;
}

std::string NumExpr::to_string() {
  return std::to_string(rep);
}

PTR(Val) NumExpr::interp(PTR(Env) env) {
  return val;
}

PTR(Expr) NumExpr::subst(std::string var, PTR(Val) new_val) {
  return THIS;
}

PTR(Expr) NumExpr::optimize() {
  return THIS;
}

bool NumExpr::has_value() {
  return true;
}

AddExpr::AddExpr(PTR(Expr) lhs, PTR(Expr) rhs) {
  this->lhs = lhs;
  this->rhs = rhs;
}

bool AddExpr::equals(PTR(Expr) other_expr) {
  PTR(AddExpr) other_add = CAST(AddExpr)(other_expr);
  if (other_add == NULL)
    return false;
  else
    return (lhs->equals(other_add->lhs)
            && rhs->equals(other_add->rhs));
}

std::string AddExpr::to_string() {
  return "(" + lhs->to_string() + "+" + rhs->to_string() + ")";
}

PTR(Val) AddExpr::interp(PTR(Env) env) {
  return lhs->interp(env)->add_to(rhs->interp(env));
}

PTR(Expr) AddExpr::subst(std::string var, PTR(Val) new_val) {
  return NEW(AddExpr)(lhs->subst(var, new_val),
                      rhs->subst(var, new_val));
}

PTR(Expr) AddExpr::optimize() {
  PTR(Expr) new_lhs = lhs->optimize();
  PTR(Expr) new_rhs = rhs->optimize();
  
  PTR(Expr) new_expr = NEW(AddExpr)(new_lhs, new_rhs);
  
  if (new_lhs->has_value() && new_rhs->has_value()) {
    try {
      return new_expr->interp(Env::empty)->to_expr();
    } catch (std::runtime_error exn) {
      // give up on folding
    }
  }
  
  return new_expr;
}

bool AddExpr::has_value() {
  return false;
}

MultExpr::MultExpr(PTR(Expr) lhs, PTR(Expr) rhs) {
  this->lhs = lhs;
  this->rhs = rhs;
}

bool MultExpr::equals(PTR(Expr) other_expr) {
  PTR(MultExpr) other_mult = CAST(MultExpr)(other_expr);
  if (other_mult == NULL)
    return false;
  else
    return (lhs->equals(other_mult->lhs)
            && rhs->equals(other_mult->rhs));
}

std::string MultExpr::to_string() {
  return "(" + lhs->to_string() + "*" + rhs->to_string() + ")";
}

PTR(Val) MultExpr::interp(PTR(Env) env) {
  return lhs->interp(env)->mult_with(rhs->interp(env));
}

PTR(Expr) MultExpr::subst(std::string var, PTR(Val) new_val)
{
  return NEW(MultExpr)(lhs->subst(var, new_val), rhs->subst(var, new_val));
}

PTR(Expr) MultExpr::optimize() {
  PTR(Expr) new_lhs = lhs->optimize();
  PTR(Expr) new_rhs = rhs->optimize();
  
  PTR(Expr) new_expr = NEW(MultExpr)(new_lhs, new_rhs);
  
  if (new_lhs->has_value() && new_rhs->has_value()) {
    try {
      return new_expr->interp(NEW(EmptyEnv)())->to_expr();
    } catch (std::runtime_error exn) {
      // give up on folding
    }
  }
  
  return new_expr;
}

bool MultExpr::has_value() {
  return false;
}

CompExpr::CompExpr(PTR(Expr) lhs, PTR(Expr) rhs) {
  this->lhs = lhs;
  this->rhs = rhs;
}

bool CompExpr::equals(PTR(Expr) other_expr) {
  PTR(CompExpr) other_comp = CAST(CompExpr)(other_expr);
  if (other_comp == NULL)
    return false;
  else
    return (lhs->equals(other_comp->lhs)
            && rhs->equals(other_comp->rhs));
}

std::string CompExpr::to_string() {
  return "(" + lhs->to_string() + "==" + rhs->to_string() + ")";
}

PTR(Val) CompExpr::interp(PTR(Env) env) {
  return NEW(BoolVal)(lhs->interp(env)->equals(rhs->interp(env)));
}

PTR(Expr) CompExpr::subst(std::string var, PTR(Val) new_val)
{
  return NEW(CompExpr)(lhs->subst(var, new_val), rhs->subst(var, new_val));
}

PTR(Expr) CompExpr::optimize() {
  PTR(Expr) new_lhs = lhs->optimize();
  PTR(Expr) new_rhs = rhs->optimize();
  
  if (new_lhs->has_value() && new_rhs->has_value()) {
    return NEW(BoolExpr)(new_lhs->interp(NEW(EmptyEnv)())->equals(new_rhs->interp(NEW(EmptyEnv)())));
  }
  
  return NEW(CompExpr)(new_lhs, new_rhs);
}

bool CompExpr::has_value() {
  return false;
}

VarExpr::VarExpr(std::string name) {
  this->name = name;
}

bool VarExpr::equals(PTR(Expr) other_expr) {
  PTR(VarExpr) other_var = CAST(VarExpr)(other_expr);
  if (other_var == NULL)
    return false;
  else
    return name == other_var->name;
}

std::string VarExpr::to_string() {
  return name;
}

PTR(Val) VarExpr::interp(PTR(Env) in_env) {
#if 0
  return in_env->lookup(name);
#else
  Env *env = in_env.get();
  
  while (1) {
    ExtendedEnv* ext_env = dynamic_cast<ExtendedEnv*>(env);
    
    if (ext_env == nullptr) {
      return in_env->lookup(name);
    } else {
      if (ext_env->name == name)
          return ext_env->val;
      env = ext_env->rest.get();
    }
  }
#endif
}

PTR(Expr) VarExpr::subst(std::string var, PTR(Val) new_val) {
  if (name == var)
    return new_val->to_expr();
  else
    return THIS;
}

PTR(Expr) VarExpr::optimize() {
  return NEW(VarExpr)(name);
}

bool VarExpr::has_value() {
  return false;
}

BoolExpr::BoolExpr(bool rep) {
  this->rep = rep;
}

bool BoolExpr::equals(PTR(Expr) other_expr) {
  PTR(BoolExpr) other_bool = CAST(BoolExpr)(other_expr);
  if (other_bool == NULL)
    return false;
  else
    return rep == other_bool->rep;
}

std::string BoolExpr::to_string() {
  return (rep ? "_true" : "_false");
}

PTR(Val) BoolExpr::interp(PTR(Env) env) {
  return NEW(BoolVal)(rep);
}

PTR(Expr) BoolExpr::subst(std::string var, PTR(Val) new_val) {
  return NEW(BoolExpr)(rep);
}

PTR(Expr) BoolExpr::optimize() {
  return NEW(BoolExpr)(rep);
}

bool BoolExpr::has_value() {
  return true;
}

LetExpr::LetExpr(std::string var, PTR(Expr) rhs, PTR(Expr) body)
{
  this->var = var;
  this->rhs = rhs;
  this->body = body;
}

bool LetExpr::equals(PTR(Expr) other_expr)
{
  PTR(LetExpr) other_let = CAST(LetExpr)(other_expr);
  if (other_let == NULL)
    return false;
  else
    return ((var == other_let->var)
            && rhs->equals(other_let->rhs)
            && body->equals(other_let->body));
}

std::string LetExpr::to_string() {
  return "(_let " + var + " = " + rhs->to_string() + " _in " + body->to_string() + ")";
}

PTR(Val) LetExpr::interp(PTR(Env) env)
{
  PTR(Val) v = rhs->interp(env);
  return body->interp(NEW(ExtendedEnv)(var, v, env));
}

PTR(Expr) LetExpr::subst(std::string repl_var, PTR(Val) new_val)
{
  if (var == repl_var)
    return NEW(LetExpr)(var,
                        rhs->subst(repl_var, new_val),
                        body);
  else
    return NEW(LetExpr)(var,
                        rhs->subst(repl_var, new_val),
                        body->subst(repl_var, new_val));
}

PTR(Expr) LetExpr::optimize() {
  PTR(Expr) new_rhs = rhs->optimize();
  
  if (new_rhs->has_value())
    return body->subst(var, new_rhs->interp(NEW(EmptyEnv)()))->optimize();
  else
    return NEW(LetExpr)(var, new_rhs, body->optimize());
}

bool LetExpr::has_value() {
  return false;
}

IfExpr::IfExpr(PTR(Expr) test_part, PTR(Expr) then_part, PTR(Expr) else_part) {
  this->test_part = test_part;
  this->then_part = then_part;
  this->else_part = else_part;
}

bool IfExpr::equals(PTR(Expr) other_expr) {
  PTR(IfExpr) other_if = CAST(IfExpr)(other_expr);
  if (other_if == NULL)
    return false;
  else
    return (test_part->equals(other_if->test_part)
            && then_part->equals(other_if->then_part)
            && else_part->equals(other_if->else_part));
}
  
std::string IfExpr::to_string() {
  return ("(_if " + test_part->to_string()
          + " _then " + then_part->to_string()
          + " _else " + else_part->to_string() + ")");
}

PTR(Val) IfExpr::interp(PTR(Env) env) {
  if (test_part->interp(env)->is_true())
    return then_part->interp(env);
  else
    return else_part->interp(env);
}

PTR(Expr) IfExpr::subst(std::string var, PTR(Val) val) {
  return NEW(IfExpr)(test_part->subst(var, val),
                     then_part->subst(var, val),
                     else_part->subst(var, val));
}

PTR(Expr) IfExpr::optimize() {
  PTR(Expr) new_test_part = test_part->optimize();
  
  if (new_test_part->has_value()) {
    PTR(Expr) branch;
    try {
      if (new_test_part->interp(NEW(EmptyEnv)())->is_true())
        branch = then_part;
      else
        branch = else_part;
    } catch (std::runtime_error exn) {
      branch = NULL;
    }
    if (branch)
      return branch->optimize();
  }

  return NEW(IfExpr)(new_test_part,
                     then_part->optimize(),
                     else_part->optimize());
}

bool IfExpr::has_value() {
  return false;
}

FunExpr::FunExpr(std::string formal_arg, PTR(Expr) body) {
  this->formal_arg = formal_arg;
  this->body = body;
}

bool FunExpr::equals(PTR(Expr) other_expr) {
  PTR(FunExpr) other_fun = CAST(FunExpr)(other_expr);
  if (other_fun == NULL)
    return false;
  else
    return ((formal_arg == other_fun->formal_arg)
            && body->equals(other_fun->body));
}

std::string FunExpr::to_string() {
  return "(_fun (" + formal_arg + ") " + body->to_string() + ")";
}
  
PTR(Val) FunExpr::interp(PTR(Env) env) {
  return NEW(FunVal)(formal_arg, body, env);
}

PTR(Expr) FunExpr::subst(std::string var, PTR(Val) val) {
  if (var == formal_arg)
    return THIS;
  else
    return NEW(FunExpr)(formal_arg, body->subst(var, val));
}

PTR(Expr) FunExpr::optimize() {
  return NEW(FunExpr)(formal_arg, body->optimize());
}

bool FunExpr::has_value() {
  return false;
}

CallExpr::CallExpr(PTR(Expr) to_be_called, PTR(Expr) actual_arg) {
  this->to_be_called = to_be_called;
  this->actual_arg = actual_arg;
}

bool CallExpr::equals(PTR(Expr) other_expr) {
  PTR(CallExpr) other_call = CAST(CallExpr)(other_expr);
  if (other_call == NULL)
    return false;
  else
    return (to_be_called->equals(other_call->to_be_called)
            && actual_arg->equals(other_call->actual_arg));
}
  
std::string CallExpr::to_string() {
  return to_be_called->to_string() + "(" + actual_arg->to_string() + ")";
}

PTR(Val) CallExpr::interp(PTR(Env) env) {
  return to_be_called->interp(env)->call(actual_arg->interp(env));
}

PTR(Expr) CallExpr::subst(std::string var, PTR(Val) val) {
  return NEW(CallExpr)(to_be_called->subst(var, val),
                       actual_arg->subst(var, val));
}

PTR(Expr) CallExpr::optimize() {
  return NEW(CallExpr)(to_be_called->optimize(),
                       actual_arg->optimize());
}

bool CallExpr::has_value() {
  return false;
}

TEST_CASE( "equals" ) {
  CHECK( (NEW(NumExpr)(1))->equals(NEW(NumExpr)(1)) );
  CHECK( ! (NEW(NumExpr)(1))->equals(NEW(NumExpr)(2)) );
  CHECK( ! (NEW(NumExpr)(1))->equals(NEW(MultExpr)(NEW(NumExpr)(2), NEW(NumExpr)(4))) );
  CHECK( (NEW(VarExpr)("x"))->equals(NEW(VarExpr)("x")) );
  CHECK( ! (NEW(VarExpr)("x"))->equals(NEW(NumExpr)(5)) );

  CHECK( (NEW(AddExpr)(NEW(NumExpr)(8), NEW(NumExpr)(9)))
         ->equals(NEW(AddExpr)(NEW(NumExpr)(8), NEW(NumExpr)(9))) );
  CHECK( ! (NEW(AddExpr)(NEW(NumExpr)(8), NEW(NumExpr)(9)))
         ->equals(NEW(AddExpr)(NEW(NumExpr)(8), NEW(NumExpr)(10))) );
  CHECK( ! (NEW(AddExpr)(NEW(NumExpr)(8), NEW(NumExpr)(9)))
         ->equals(NEW(AddExpr)(NEW(NumExpr)(10), NEW(NumExpr)(9))) );
  CHECK( ! (NEW(AddExpr)(NEW(NumExpr)(8), NEW(NumExpr)(9)))
         ->equals(NEW(NumExpr)(8)) );

  CHECK( (NEW(MultExpr)(NEW(NumExpr)(8), NEW(NumExpr)(9)))
         ->equals(NEW(MultExpr)(NEW(NumExpr)(8), NEW(NumExpr)(9))) );
  CHECK( ! (NEW(MultExpr)(NEW(NumExpr)(8), NEW(NumExpr)(9)))
         ->equals(NEW(AddExpr)(NEW(NumExpr)(8), NEW(NumExpr)(9))) );
  CHECK( (NEW(CompExpr)(NEW(NumExpr)(8), NEW(NumExpr)(9)))
         ->equals(NEW(CompExpr)(NEW(NumExpr)(8), NEW(NumExpr)(9))) );
  CHECK( ! (NEW(CompExpr)(NEW(NumExpr)(8), NEW(NumExpr)(9)))
         ->equals(NEW(AddExpr)(NEW(NumExpr)(8), NEW(NumExpr)(9))) );

  CHECK( (NEW(LetExpr)("x", NEW(NumExpr)(5), NEW(NumExpr)(6)))
         ->equals(NEW(LetExpr)("x", NEW(NumExpr)(5), NEW(NumExpr)(6))) );
  CHECK( ! (NEW(LetExpr)("x", NEW(NumExpr)(5), NEW(NumExpr)(6)))
         ->equals(NEW(LetExpr)("y", NEW(NumExpr)(5), NEW(NumExpr)(6))) );
  CHECK( ! (NEW(LetExpr)("x", NEW(NumExpr)(5), NEW(NumExpr)(6)))
         ->equals(NEW(LetExpr)("x", NEW(NumExpr)(7), NEW(NumExpr)(6))) );
  CHECK( ! (NEW(LetExpr)("x", NEW(NumExpr)(5), NEW(NumExpr)(6)))
         ->equals(NEW(LetExpr)("x", NEW(NumExpr)(5), NEW(NumExpr)(7))) );
  CHECK( ! (NEW(LetExpr)("x", NEW(NumExpr)(5), NEW(NumExpr)(6)))
         ->equals(NEW(NumExpr)(5)) );

  CHECK( (NEW(BoolExpr)(true))->equals(NEW(BoolExpr)(true)) );
  CHECK( ! (NEW(BoolExpr)(true))->equals(NEW(BoolExpr)(false)) );
  CHECK( ! (NEW(BoolExpr)(true))->equals(NEW(NumExpr)(2)) );

  CHECK( (NEW(IfExpr)(NEW(NumExpr)(1),
                      NEW(AddExpr)(NEW(NumExpr)(2), NEW(NumExpr)(2)),
                      NEW(BoolExpr)(true)))
        ->equals(NEW(IfExpr)(NEW(NumExpr)(1),
                             NEW(AddExpr)(NEW(NumExpr)(2), NEW(NumExpr)(2)),
                             NEW(BoolExpr)(true))) );
  CHECK( ! (NEW(IfExpr)(NEW(NumExpr)(1),
                        NEW(AddExpr)(NEW(NumExpr)(2), NEW(NumExpr)(2)),
                        NEW(BoolExpr)(true)))
         ->equals(NEW(BoolExpr)(false)) );
  
  CHECK( (NEW(FunExpr)("x", NEW(VarExpr)("x")))
         ->equals(NEW(FunExpr)("x", NEW(VarExpr)("x"))) );
  CHECK( ! (NEW(FunExpr)("x", NEW(VarExpr)("x")))
         ->equals(NEW(FunExpr)("y", NEW(VarExpr)("x"))) );
  CHECK( ! (NEW(FunExpr)("x", NEW(VarExpr)("x")))
         ->equals(NEW(FunExpr)("x", NEW(VarExpr)("y"))) );
  CHECK( ! (NEW(FunExpr)("x", NEW(VarExpr)("x")))
         ->equals(NEW(NumExpr)(7)) );
  
  CHECK( (NEW(CallExpr)(NEW(NumExpr)(1), NEW(NumExpr)(2)))
         ->equals(NEW(CallExpr)(NEW(NumExpr)(1), NEW(NumExpr)(2))) );
  CHECK( ! (NEW(CallExpr)(NEW(NumExpr)(1), NEW(NumExpr)(2)))
         ->equals(NEW(CallExpr)(NEW(NumExpr)(3), NEW(NumExpr)(2))) );
  CHECK( ! (NEW(CallExpr)(NEW(NumExpr)(1), NEW(NumExpr)(2)))
         ->equals(NEW(CallExpr)(NEW(NumExpr)(1), NEW(NumExpr)(3))) );
  CHECK( ! (NEW(CallExpr)(NEW(NumExpr)(1), NEW(NumExpr)(2)))
         ->equals(NEW(NumExpr)(1)) );
}

TEST_CASE( "Expr to_string" ) {
  CHECK( (NEW(NumExpr)(10))->to_string() == "10" );
  CHECK( (NEW(BoolExpr)(true))->to_string() == "_true" );
  CHECK( (NEW(BoolExpr)(false))->to_string() == "_false" );
  CHECK( (NEW(VarExpr)("z"))->to_string() == "z" );
  CHECK( (NEW(AddExpr)(NEW(NumExpr)(10), NEW(NumExpr)(8)))->to_string() == "(10+8)" );
  CHECK( (NEW(MultExpr)(NEW(NumExpr)(10), NEW(NumExpr)(8)))->to_string() == "(10*8)" );
  CHECK( (NEW(CompExpr)(NEW(NumExpr)(10), NEW(NumExpr)(8)))->to_string() == "(10==8)" );
  CHECK( (NEW(LetExpr)("x", NEW(NumExpr)(10), NEW(NumExpr)(8)))->to_string()
         == "(_let x = 10 _in 8)" );
  CHECK( (NEW(IfExpr)(NEW(NumExpr)(1), NEW(NumExpr)(2), NEW(NumExpr)(3)))->to_string()
         == "(_if 1 _then 2 _else 3)" );
  CHECK( (NEW(FunExpr)("x", NEW(NumExpr)(5)))->to_string() == "(_fun (x) 5)" );
  CHECK( (NEW(CallExpr)(NEW(VarExpr)("x"), NEW(VarExpr)("y")))->to_string() == "x(y)" );
}

TEST_CASE( "interp" ) {
  CHECK( (NEW(NumExpr)(10))->interp(Env::empty)->equals(NEW(NumVal)(10)) );
  CHECK_THROWS_WITH( (NEW(VarExpr)("fish"))->interp(Env::empty),
                     "free variable: fish");
  CHECK( (NEW(VarExpr)("fish"))->interp(NEW(ExtendedEnv)("fish", NEW(NumVal)(12), Env::empty))
         ->equals(NEW(NumVal)(12)) );
  CHECK( (NEW(AddExpr)(NEW(NumExpr)(3), NEW(NumExpr)(2)))->interp(Env::empty)
         ->equals(NEW(NumVal)(5)) );
  CHECK( (NEW(MultExpr)(NEW(NumExpr)(3), NEW(NumExpr)(2)))->interp(Env::empty)
         ->equals(NEW(NumVal)(6)) );
  CHECK( (NEW(CompExpr)(NEW(NumExpr)(3), NEW(NumExpr)(3)))->interp(Env::empty)
         ->equals(NEW(BoolVal)(true)) );
  CHECK( (NEW(LetExpr)("x", NEW(NumExpr)(5), NEW(VarExpr)("x")))->interp(Env::empty)
         ->equals(NEW(NumVal)(5)) );
  CHECK( (NEW(BoolExpr)(true))->interp(Env::empty)->equals(NEW(BoolVal)(true)) );
  CHECK( (NEW(IfExpr)(NEW(BoolExpr)(true),
                      NEW(NumExpr)(3),
                      NEW(NumExpr)(4)))->interp(Env::empty)
         ->equals(NEW(NumVal)(3)) );
  CHECK( (NEW(IfExpr)(NEW(BoolExpr)(false),
                      NEW(NumExpr)(3),
                      NEW(NumExpr)(4)))->interp(Env::empty)
         ->equals(NEW(NumVal)(4)) );
  
  CHECK( (NEW(AddExpr)(NEW(VarExpr)("x"), NEW(VarExpr)("y")))
        ->interp(NEW(ExtendedEnv)("x", NEW(NumVal)(4),
                                  NEW(ExtendedEnv)("y", NEW(NumVal)(3),
                                                   Env::empty)))
        ->equals(NEW(NumVal)(7)) );
  
}

TEST_CASE( "subst") {
  CHECK( (NEW(NumExpr)(10))->subst("dog", NEW(NumVal)(3))
         ->equals(NEW(NumExpr)(10)) );
  CHECK( (NEW(BoolExpr)(true))->subst("dog", NEW(NumVal)(3))
         ->equals(NEW(BoolExpr)(true)) );
  CHECK( (NEW(VarExpr)("fish"))->subst("dog", NEW(NumVal)(3))
         ->equals(NEW(VarExpr)("fish")) );
  CHECK( (NEW(VarExpr)("dog"))->subst("dog", NEW(NumVal)(3) )
         ->equals(NEW(NumExpr)(3)) );
  CHECK( (NEW(VarExpr)("dog"))->subst("dog", NEW(BoolVal)(true) )
         ->equals(NEW(BoolExpr)(true)) );
  CHECK( (NEW(AddExpr)(NEW(NumExpr)(2), NEW(VarExpr)("dog")))->subst("dog", NEW(NumVal)(3))
         ->equals(NEW(AddExpr)(NEW(NumExpr)(2), NEW(NumExpr)(3))) );
  CHECK( (NEW(MultExpr)(NEW(NumExpr)(2), NEW(VarExpr)("dog")))->subst("dog", NEW(NumVal)(3))
         ->equals(NEW(MultExpr)(NEW(NumExpr)(2), NEW(NumExpr)(3))) );
  CHECK( (NEW(CompExpr)(NEW(NumExpr)(2), NEW(VarExpr)("dog")))->subst("dog", NEW(NumVal)(3))
         ->equals(NEW(CompExpr)(NEW(NumExpr)(2), NEW(NumExpr)(3))) );
  CHECK( (NEW(LetExpr)("x", NEW(VarExpr)("x"), NEW(VarExpr)("x")))->subst("x", NEW(NumVal)(5))
         ->equals(NEW(LetExpr)("x", NEW(NumExpr)(5), NEW(VarExpr)("x"))) );
  CHECK( (NEW(LetExpr)("y", NEW(VarExpr)("x"), NEW(VarExpr)("x")))->subst("x", NEW(NumVal)(5))
         ->equals(NEW(LetExpr)("y", NEW(NumExpr)(5), NEW(NumExpr)(5))) );
  CHECK( (NEW(IfExpr)(NEW(CompExpr)(NEW(NumExpr)(1), NEW(VarExpr)("y")),
                      NEW(AddExpr)(NEW(NumExpr)(1), NEW(VarExpr)("y")),
                      NEW(MultExpr)(NEW(NumExpr)(1), NEW(VarExpr)("y"))))
         ->subst("y", NEW(NumVal)(8))
         ->equals(NEW(IfExpr)(NEW(CompExpr)(NEW(NumExpr)(1), NEW(NumExpr)(8)),
                              NEW(AddExpr)(NEW(NumExpr)(1), NEW(NumExpr)(8)),
                              NEW(MultExpr)(NEW(NumExpr)(1), NEW(NumExpr)(8)))) );
  CHECK( (NEW(FunExpr)("x", NEW(VarExpr)("x")))->subst("x" , NEW(NumVal)(10))
         ->equals(NEW(FunExpr)("x", NEW(VarExpr)("x"))) );
  CHECK( (NEW(FunExpr)("x", NEW(VarExpr)("y")))->subst("y" , NEW(NumVal)(10))
         ->equals(NEW(FunExpr)("x", NEW(NumExpr)(10))) );
  CHECK( (NEW(CallExpr)(NEW(VarExpr)("x"), NEW(AddExpr)(NEW(VarExpr)("x"), NEW(NumExpr)(1))))
         ->subst("x", NEW(NumVal)(10))
         ->equals(NEW(CallExpr)(NEW(NumExpr)(10), NEW(AddExpr)(NEW(NumExpr)(10), NEW(NumExpr)(1)))) );
}

TEST_CASE( "optimize" ) {
  CHECK( (NEW(NumExpr)(10))->optimize()->equals(NEW(NumExpr)(10)) );
  CHECK( (NEW(VarExpr)("fish"))->optimize()->equals(NEW(VarExpr)("fish")) );
  CHECK( (NEW(AddExpr)(NEW(NumExpr)(3), NEW(NumExpr)(2)))->optimize()
         ->equals(NEW(NumExpr)(5)) );
  CHECK( (NEW(AddExpr)(NEW(NumExpr)(3), NEW(VarExpr)("x")))->optimize()
         ->equals(NEW(AddExpr)(NEW(NumExpr)(3), NEW(VarExpr)("x"))) );
  CHECK( (NEW(AddExpr)(NEW(NumExpr)(3), NEW(BoolExpr)(true)))->optimize()
         ->equals(NEW(AddExpr)(NEW(NumExpr)(3), NEW(BoolExpr)(true))) );
  CHECK( (NEW(MultExpr)(NEW(NumExpr)(3), NEW(NumExpr)(2)))->optimize()
         ->equals(NEW(NumExpr)(6)) );
  CHECK( (NEW(MultExpr)(NEW(NumExpr)(3), NEW(BoolExpr)(true)))->optimize()
         ->equals(NEW(MultExpr)(NEW(NumExpr)(3), NEW(BoolExpr)(true))) );
  CHECK( (NEW(MultExpr)(NEW(NumExpr)(3), NEW(VarExpr)("x")))->optimize()
         ->equals(NEW(MultExpr)(NEW(NumExpr)(3), NEW(VarExpr)("x"))) );
  CHECK( (NEW(CompExpr)(NEW(NumExpr)(3), NEW(NumExpr)(3)))->optimize()
         ->equals(NEW(BoolExpr)(true)) );
  CHECK( (NEW(CompExpr)(NEW(NumExpr)(3), NEW(VarExpr)("x")))->optimize()
         ->equals(NEW(CompExpr)(NEW(NumExpr)(3), NEW(VarExpr)("x"))) );
  CHECK( (NEW(CompExpr)(NEW(NumExpr)(3), NEW(BoolExpr)(false)))->optimize()
         ->equals(NEW(BoolExpr)(false)) );
  CHECK( (NEW(LetExpr)("x", NEW(NumExpr)(5), NEW(VarExpr)("x")))->optimize()
         ->equals(NEW(NumExpr)(5)) );
  CHECK( (NEW(LetExpr)("x", NEW(VarExpr)("y"), NEW(VarExpr)("x")))->optimize()
         ->equals(NEW(LetExpr)("x", NEW(VarExpr)("y"), NEW(VarExpr)("x"))) );
  CHECK( (NEW(BoolExpr)(true))->optimize()->equals(NEW(BoolExpr)(true)) );
  CHECK( (NEW(IfExpr)(NEW(BoolExpr)(true),
                      NEW(NumExpr)(3),
                      NEW(NumExpr)(4)))->optimize()
         ->equals(NEW(NumExpr)(3)) );
  CHECK( (NEW(IfExpr)(NEW(BoolExpr)(false),
                      NEW(NumExpr)(3),
                      NEW(NumExpr)(4)))->optimize()
         ->equals(NEW(NumExpr)(4)) );
  CHECK( (NEW(IfExpr)(NEW(NumExpr)(8),
                      NEW(NumExpr)(3),
                      NEW(NumExpr)(4)))->optimize()
         ->equals(NEW(IfExpr)(NEW(NumExpr)(8),
                              NEW(NumExpr)(3),
                              NEW(NumExpr)(4))) );
  CHECK( (NEW(FunExpr)("x", NEW(AddExpr)(NEW(NumExpr)(1), NEW(NumExpr)(2))))->optimize()
         ->equals(NEW(FunExpr)("x", NEW(NumExpr)(3))) );
  CHECK( (NEW(CallExpr)(NEW(MultExpr)(NEW(NumExpr)(3), NEW(NumExpr)(4)),
                        NEW(AddExpr)(NEW(NumExpr)(1), NEW(NumExpr)(2))))->optimize()
         ->equals(NEW(CallExpr)(NEW(NumExpr)(12), NEW(NumExpr)(3))) );
}

TEST_CASE( "has_value" ) {
  CHECK( (NEW(NumExpr)(10))->has_value() );
  CHECK( ! (NEW(VarExpr)("fish"))->has_value() );
  CHECK( ! (NEW(AddExpr)(NEW(NumExpr)(3), NEW(NumExpr)(2)))->has_value() );
  CHECK( ! (NEW(MultExpr)(NEW(NumExpr)(3), NEW(NumExpr)(2)))->has_value() );
  CHECK( ! (NEW(CompExpr)(NEW(NumExpr)(3), NEW(NumExpr)(2)))->has_value() );
  CHECK( ! (NEW(LetExpr)("x", NEW(NumExpr)(5), NEW(VarExpr)("x")))->has_value() );
  CHECK( (NEW(BoolExpr)(true))->has_value() );
  CHECK( ! (NEW(IfExpr)(NEW(BoolExpr)(true),
                        NEW(NumExpr)(3),
                        NEW(NumExpr)(4)))->has_value() );
  CHECK( !(NEW(FunExpr)("x", NEW(VarExpr)("x")))->has_value() );
  CHECK( !(NEW(CallExpr)(NEW(VarExpr)("x"), NEW(VarExpr)("x")))->has_value() );
}

TEST_CASE( "interp stress" ) {
  PTR(Expr) fib_call = parse_str("_let fib = _fun (fib)"
                                 "               _fun (x)"
                                 "                  _if x == 0"
                                 "                  _then 1"
                                 "                  _else _if x == 2 + -1"
                                 "                  _then 1"
                                 "                  _else fib(fib)(x + -1)"
                                 "                        + fib(fib)(x + -2)"
                                 "  _in fib(fib)(10)");
  CHECK( fib_call->interp(Env::empty)->equals(NEW(NumVal)(89)) );
}
