#ifndef expr_hpp
#define expr_hpp

#include "pointer.hpp"

#include <string>

class Val;
class Env;

class Expr ENABLE_THIS(Expr) {
public:
  virtual bool equals(PTR(Expr) other_expr) = 0;

  // Print an expression in a way that can be
  // parsed back in
  virtual std::string to_string() = 0;

  // To compute the number value of an expression,
  // assuming that all variables are 0
  virtual PTR(Val) interp(PTR(Env) env) = 0;
  
  // To substitute a value (which must be converted to an expression)
  // in place of a variable within the expression
  virtual PTR(Expr) subst(std::string var, PTR(Val) val) = 0;
  
  // To simplify the program to produce another one that
  // evaluates the same; must not error and must not make
  // a new program that runs slower than the original
  virtual PTR(Expr) optimize() = 0;
  
  // Assuming that this is an optimized expression,
  // can it provide a value (without erroring)?
  virtual bool has_value() = 0;
};

class NumExpr : public Expr {
public:
  int rep;
  PTR(Val) val;

  NumExpr(int rep);
  bool equals(PTR(Expr) other_expr);

  std::string to_string();

  PTR(Val) interp(PTR(Env) env);
  PTR(Expr) subst(std::string var, PTR(Val) val);
  
  PTR(Expr) optimize();
  bool has_value();
};

class AddExpr : public Expr {
public:
  PTR(Expr) lhs;
  PTR(Expr) rhs;

  AddExpr(PTR(Expr) lhs, PTR(Expr) rhs);
  bool equals(PTR(Expr) other_expr);

  std::string to_string();

  PTR(Val) interp(PTR(Env) env);
  PTR(Expr) subst(std::string var, PTR(Val) val);

  PTR(Expr) optimize();
  bool has_value();
};

class MultExpr : public Expr {
public:
  PTR(Expr) lhs;
  PTR(Expr) rhs;

  MultExpr(PTR(Expr) lhs, PTR(Expr) rhs);
  bool equals(PTR(Expr) other_expr);

  std::string to_string();

  PTR(Val) interp(PTR(Env) env);
  PTR(Expr) subst(std::string var, PTR(Val) val);

  PTR(Expr) optimize();
  bool has_value();
};

class CompExpr : public Expr {
public:
  PTR(Expr) lhs;
  PTR(Expr) rhs;

  CompExpr(PTR(Expr) lhs, PTR(Expr) rhs);
  bool equals(PTR(Expr) other_expr);

  std::string to_string();

  PTR(Val) interp(PTR(Env) env);
  PTR(Expr) subst(std::string var, PTR(Val) val);

  PTR(Expr) optimize();
  bool has_value();
};

class VarExpr : public Expr {
public:
  std::string name;

  VarExpr(std::string name);
  bool equals(PTR(Expr) other_expr);

  std::string to_string();

  PTR(Val) interp(PTR(Env) env);
  PTR(Expr) subst(std::string var, PTR(Val) val);

  PTR(Expr) optimize();
  bool has_value();
};

class LetExpr : public Expr {
public:
  std::string var;
  PTR(Expr) rhs;
  PTR(Expr) body;

  LetExpr(std::string var, PTR(Expr) rhs, PTR(Expr) body);
  bool equals(PTR(Expr) other_expr);

  std::string to_string();

  PTR(Val) interp(PTR(Env) env);
  PTR(Expr) subst(std::string var, PTR(Val) val);

  PTR(Expr) optimize();
  bool has_value();
};

class BoolExpr : public Expr {
public:
  bool rep;
  
  BoolExpr(bool rep);
  bool equals(PTR(Expr) other_expr);
  
  std::string to_string();

  PTR(Val) interp(PTR(Env) env);
  PTR(Expr) subst(std::string var, PTR(Val) val);

  PTR(Expr) optimize();
  bool has_value();
};

class IfExpr : public Expr {
public:
  PTR(Expr) test_part;
  PTR(Expr) then_part;
  PTR(Expr) else_part;

  IfExpr(PTR(Expr) test_part, PTR(Expr) then_part, PTR(Expr) else_part);
  bool equals(PTR(Expr) other_expr);
  
  std::string to_string();

  PTR(Val) interp(PTR(Env) env);
  PTR(Expr) subst(std::string var, PTR(Val) val);

  PTR(Expr) optimize();
  bool has_value();
};

class FunExpr : public Expr {
public:
  std::string formal_arg;
  PTR(Expr) body;

  FunExpr(std::string formal_arg, PTR(Expr) body);
  bool equals(PTR(Expr) other_expr);
  
  std::string to_string();

  PTR(Val) interp(PTR(Env) env);
  PTR(Expr) subst(std::string var, PTR(Val) val);

  PTR(Expr) optimize();
  bool has_value();
};

class CallExpr : public Expr {
public:
  PTR(Expr) to_be_called;
  PTR(Expr) actual_arg;

  std::string to_string();

  CallExpr(PTR(Expr) to_be_called, PTR(Expr) actual_arg);
  bool equals(PTR(Expr) other_expr);
  
  PTR(Val) interp(PTR(Env) env);
  PTR(Expr) subst(std::string var, PTR(Val) val);

  PTR(Expr) optimize();
  bool has_value();
};

#endif /* expr_hpp */
