#include <iostream>
#include <sstream>
#include <cassert>
#include "parse.hpp"
#include "expr.hpp"
#include "catch.hpp"

static PTR(Expr) parse_expr(std::istream &in);
static PTR(Expr) parse_comparg(std::istream &in);
static PTR(Expr) parse_addend(std::istream &in);
static PTR(Expr) parse_multicand(std::istream &in);
static PTR(Expr) parse_inner(std::istream &in);
static PTR(Expr) parse_number(std::istream &in, char start_ch);
static PTR(Expr) parse_variable(std::istream &in);
static PTR(Expr) parse_parenthesized(std::istream &in);
static PTR(Expr) parse_let_form(std::istream &in);
static PTR(Expr) parse_if_form(std::istream &in);
static PTR(Expr) parse_fun_form(std::istream &in);
static std::string parse_keyword(std::istream &in);
static std::string parse_alphabetic(std::istream &in, std::string prefix);
static void parse_expected_keyword(std::istream &in, std::string kw, std::string for_kw);
static char peek_after_spaces(std::istream &in);
static void consume(std::istream &in, char expect);

// Take an input stream that contains an expression,
// and returns the parsed representation of that expression.
// Throws `runtime_error` for parse errors.
PTR(Expr) parse(std::istream &in) {
  PTR(Expr) expr = parse_expr(in);
  
  char next_ch = in.get();
  if (!in.eof())
    throw std::runtime_error((std::string)"expected end of file at " + next_ch);
  
  return expr;
}

// Takes an input stream that starts with an expression,
// consuming the largest initial expression possible.
static PTR(Expr) parse_expr(std::istream &in) {
  PTR(Expr) expr = parse_comparg(in);
  
  if (peek_after_spaces(in) == '=') {
    consume(in, '=');

    char next_ch = peek_after_spaces(in);
    if (next_ch != '=')
      throw std::runtime_error((std::string)"expected `=` after `=` instead of " + next_ch);
    consume(in, '=');
    
    PTR(Expr) rhs = parse_expr(in);
    return NEW(CompExpr)(expr, rhs);
  } else
    return expr;
}

// Takes an input stream that starts with a comparg,
// consuming the largest initial expression possible,
// where a comparg is something that does not have
// `==` except within nested expressions (like parentheses).
static PTR(Expr) parse_comparg(std::istream &in) {
  PTR(Expr) expr = parse_addend(in);
  
  if (peek_after_spaces(in) == '+') {
    consume(in, '+');
    PTR(Expr) rhs = parse_comparg(in);
    return NEW(AddExpr)(expr, rhs);
  } else
    return expr;
}

// Takes an input stream that starts with an addend,
// consuming the largest initial addend possible, where
// an addend is an expression that does not have `+`
// except within nested expressions (like parentheses).
static PTR(Expr) parse_addend(std::istream &in) {
  PTR(Expr) expr = parse_multicand(in);
  
  if (peek_after_spaces(in) == '*') {
    consume(in, '*');
    PTR(Expr) rhs = parse_addend(in);
    return NEW(MultExpr)(expr, rhs);
  } else
    return expr;
}

// Takes an input stream that starts with an addend,
// consuming the largest initial addend possible, where
// an addend is an expression that does not have `+`
// except within nested expressions (like parentheses).
static PTR(Expr) parse_multicand(std::istream &in) {
  PTR(Expr) expr = parse_inner(in);
  
  while (peek_after_spaces(in) == '(') {
    consume(in, '(');
    PTR(Expr) actual_arg = parse_expr(in);

    if (peek_after_spaces(in) != ')')
      throw std::runtime_error("expected a close parenthesis for a function call");
    consume(in, ')');
    
    expr = NEW(CallExpr)(expr, actual_arg);
  }
  
  return expr;
}

// Parses something with no immediate `==`, `+`, or `*` from `in`.
static PTR(Expr) parse_inner(std::istream &in) {
  char next_ch = peek_after_spaces(in);
  
  if (next_ch == '(')
    return parse_parenthesized(in);
  else if (isdigit(next_ch) || (next_ch == '-'))
    return parse_number(in, next_ch);
  else if (isalpha(next_ch))
    return parse_variable(in);
  else if (next_ch == '_') {
    std::string keyword = parse_keyword(in);
    if (keyword == "_true")
      return NEW(BoolExpr)(true);
    else if (keyword == "_false")
      return NEW(BoolExpr)(false);
    else if (keyword == "_let")
      return parse_let_form(in);
    else if (keyword == "_if")
      return parse_if_form(in);
    else if (keyword == "_fun")
      return parse_fun_form(in);
    else
      throw std::runtime_error("unexpected keyword " + keyword);
  } else
    throw std::runtime_error((std::string)"unexpected input at " + next_ch);
}

static PTR(Expr) parse_parenthesized(std::istream &in) {
  consume(in, '(');

  PTR(Expr) expr = parse_expr(in);

  if (peek_after_spaces(in) == ')')
    consume(in, ')');
  else
    throw std::runtime_error("expected a close parenthesis");
  
  return expr;
}

static PTR(Expr) parse_let_form(std::istream &in) {
  char next_ch = peek_after_spaces(in);
  if (!isalpha(next_ch))
    throw std::runtime_error("expected a name after `_let`");
  std::string var = parse_alphabetic(in, "");
  
  if (peek_after_spaces(in) != '=')
    throw std::runtime_error("expected a `=` for `_let` binding");
  consume(in, '=');

  PTR(Expr) rhs = parse_expr(in);
  
  parse_expected_keyword(in, "_in", "_let");
  
  PTR(Expr) body = parse_expr(in);
  
  return NEW(LetExpr)(var, rhs, body);
}

static PTR(Expr) parse_if_form(std::istream &in) {
  PTR(Expr) test_part = parse_expr(in);
  
  parse_expected_keyword(in, "_then", "_if");

  PTR(Expr) then_part = parse_expr(in);

  parse_expected_keyword(in, "_else", "_if");
  
  PTR(Expr) else_part = parse_expr(in);

  return NEW(IfExpr)(test_part, then_part, else_part);
}

static PTR(Expr) parse_fun_form(std::istream &in) {
  if (peek_after_spaces(in) != '(')
    throw std::runtime_error("expected `(` after `_fun`");
  consume(in, '(');

  char next_ch = peek_after_spaces(in);
  if (!isalpha(next_ch))
    throw std::runtime_error("expected a name after `(` for `_fun`");
  std::string formal_arg = parse_alphabetic(in, "");

  if (peek_after_spaces(in) != ')')
    throw std::runtime_error("expected `)` after argument for `_fun`");
  consume(in, ')');

  PTR(Expr) body = parse_expr(in);
  
  return NEW(FunExpr)(formal_arg, body);
}

static void parse_expected_keyword(std::istream &in, std::string kw, std::string for_kw) {
  if (peek_after_spaces(in) != '_')
    throw std::runtime_error("expected `" + kw + "` for `" + for_kw + "`");
  consume(in, '_');
  std::string name = parse_alphabetic(in, "_");
  if (name != kw)
    throw std::runtime_error("expected `" + kw + "` for `" + for_kw + "`");
}

// Parses a number, assuming that `in` starts with a digit.
static PTR(Expr) parse_number(std::istream &in, char start_ch) {
  int num = 0;

  if (start_ch == '-')
    consume(in, '-');
  
  char next_ch = in.peek();
  if (!isdigit(next_ch))
    throw std::runtime_error("expected a digit after `-`");
  
  in >> num;

  if (start_ch == '-')
    num = -num;

  return NEW(NumExpr)(num);
}

// Parses an expression, assuming that `in` starts with a
// letter.
static PTR(Expr) parse_variable(std::istream &in) {
  return NEW(VarExpr)(parse_alphabetic(in, ""));
}

// Parses an expression, assuming that `in` starts with a
// letter.
static std::string parse_keyword(std::istream &in) {
  consume(in, '_');
  return parse_alphabetic(in, "_");
}

// Parses an expression, assuming that `in` starts with a
// letter.
static std::string parse_alphabetic(std::istream &in, std::string prefix) {
  std::string name = prefix;
  while (1) {
    char next_ch = in.peek();
    if (!isalpha(next_ch))
      break;
    consume(in, next_ch);
    name += next_ch;
  }
  return name;
}

// Like in.peek(), but consume an whitespace at the
// start of `in`
static char peek_after_spaces(std::istream &in) {
  char next_ch;
  while (1) {
    next_ch = in.peek();
    if (!isspace(next_ch))
      break;
    consume(in, next_ch);
  }
  return next_ch;
}

static void consume(std::istream &in, char expect) {
  char next_ch = in.get();
  assert(next_ch == expect);
}

PTR(Expr) parse_str(std::string s) {
  std::istringstream in(s);
  return parse(in);
}

/* for tests */
static std::string parse_str_error(std::string s) {
  std::istringstream in(s);
  try {
    (void)parse(in);
    return "";
  } catch (std::runtime_error exn) {
    return exn.what();
  }
}

TEST_CASE( "parser" ) {
  PTR(Expr) ten_plus_one = NEW(AddExpr)(NEW(NumExpr)(10), NEW(NumExpr)(1));
  
  CHECK ( parse_str_error(" ( 1 ") == "expected a close parenthesis" );

  CHECK( parse_str("10")->equals(NEW(NumExpr)(10)) );
  CHECK( parse_str("-10")->equals(NEW(NumExpr)(-10)) );
  CHECK( parse_str("(10)")->equals(NEW(NumExpr)(10)) );
  CHECK( parse_str("10+1")->equals(ten_plus_one) );
  CHECK( parse_str("(10+1)")->equals(ten_plus_one) );
  CHECK( parse_str("(10)+1")->equals(ten_plus_one) );
  CHECK( parse_str("10+(1)")->equals(ten_plus_one) );
  CHECK( parse_str("1+2*3")->equals(NEW(AddExpr)(NEW(NumExpr)(1),
                                                 NEW(MultExpr)(NEW(NumExpr)(2), NEW(NumExpr)(3)))) );
  CHECK( parse_str("1*2+3")->equals(NEW(AddExpr)(NEW(MultExpr)(NEW(NumExpr)(1), NEW(NumExpr)(2)),
                                                 NEW(NumExpr)(3))) );
  CHECK( parse_str("4*2*3")->equals(NEW(MultExpr)(NEW(NumExpr)(4),
                                                  NEW(MultExpr)(NEW(NumExpr)(2), NEW(NumExpr)(3)))) );
  CHECK( parse_str("4+2+3")->equals(NEW(AddExpr)(NEW(NumExpr)(4),
                                                 NEW(AddExpr)(NEW(NumExpr)(2), NEW(NumExpr)(3)))) );
  CHECK( parse_str("4*(2+3)")->equals(NEW(MultExpr)(NEW(NumExpr)(4),
                                                    NEW(AddExpr)(NEW(NumExpr)(2), NEW(NumExpr)(3)))) );
  CHECK( parse_str("(2+3)*4")->equals(NEW(MultExpr)(NEW(AddExpr)(NEW(NumExpr)(2), NEW(NumExpr)(3)),
                                                    NEW(NumExpr)(4))) );
  CHECK( parse_str("1 == 2")->equals(NEW(CompExpr)(NEW(NumExpr)(1), NEW(NumExpr)(2))) );
  CHECK( parse_str("1+2 == 2+1")->equals(NEW(CompExpr)(NEW(AddExpr)(NEW(NumExpr)(1), NEW(NumExpr)(2)),
                                                       NEW(AddExpr)(NEW(NumExpr)(2), NEW(NumExpr)(1)))) );
  CHECK( parse_str("xyz")->equals(NEW(VarExpr)("xyz")) );
  CHECK( parse_str("xyz+1")->equals(NEW(AddExpr)(NEW(VarExpr)("xyz"), NEW(NumExpr)(1))) );

  CHECK ( parse_str_error("!") == "unexpected input at !" );
  CHECK ( parse_str_error("(1") == "expected a close parenthesis" );
  CHECK ( parse_str_error("1 =! 3") == "expected `=` after `=` instead of !" );
  CHECK ( parse_str_error("- 1") == "expected a digit after `-`" );

  CHECK( parse_str(" 10 ")->equals(NEW(NumExpr)(10)) );
  CHECK( parse_str(" (  10 ) ")->equals(NEW(NumExpr)(10)) );
  CHECK( parse_str(" 10  + 1")->equals(ten_plus_one) );
  CHECK( parse_str(" ( 10 + 1 ) ")->equals(ten_plus_one) );
  CHECK( parse_str(" 11 * ( 10 + 1 ) ")->equals(NEW(MultExpr)(NEW(NumExpr)(11),
                                                              ten_plus_one)) );
  CHECK( parse_str(" ( 11 * 10 ) + 1 ")
         ->equals(NEW(AddExpr)(NEW(MultExpr)(NEW(NumExpr)(11), NEW(NumExpr)(10)),
                               NEW(NumExpr) (1))) );
  CHECK( parse_str(" 1 + 2 * 3 ")
         ->equals(NEW(AddExpr)(NEW(NumExpr)(1),
                               NEW(MultExpr)(NEW(NumExpr)(2), NEW(NumExpr)(3)))) );

  CHECK( parse_str_error(" ! ") == "unexpected input at !" );
  CHECK( parse_str_error(" ( 1 ") == "expected a close parenthesis" );
  CHECK( parse_str_error(" 1 )") == "expected end of file at )" );
  
  CHECK( parse_str("_true")->equals(NEW(BoolExpr)(true)) );
  CHECK( parse_str("_false")->equals(NEW(BoolExpr)(false)) );
  
  CHECK( parse_str_error("_maybe ") == "unexpected keyword _maybe" );
  
  CHECK( parse_str("_let x = 2*3+4 _in x+x")
         ->equals(NEW(LetExpr)("x",
                               NEW(AddExpr)(NEW(MultExpr)(NEW(NumExpr)(2), NEW(NumExpr)(3)),
                                            NEW(NumExpr)(4)),
                               NEW(AddExpr)(NEW(VarExpr)("x"), NEW(VarExpr)("x")))) );
  
  CHECK ( parse_str_error(" _let ! ")
          == "expected a name after `_let`" );
  CHECK ( parse_str_error(" _let x ! ")
          == "expected a `=` for `_let` binding" );
  CHECK ( parse_str_error(" _let x = 4 ! ")
          == "expected `_in` for `_let`" );
  CHECK ( parse_str_error(" _let x = 4 _other ")
          == "expected `_in` for `_let`" );
  CHECK ( parse_str_error(" _let x = 4 _in ")
          == "unexpected input at \377" );
  
  CHECK( parse_str("_if 1 _then 2+2 _else _true")
         ->equals(NEW(IfExpr)(NEW(NumExpr)(1),
                              NEW(AddExpr)(NEW(NumExpr)(2), NEW(NumExpr)(2)),
                              NEW(BoolExpr)(true))) );
  
  CHECK( parse_str("_fun (x) x + 1")
         ->equals(NEW(FunExpr)("x", NEW(AddExpr)(NEW(VarExpr)("x"), NEW(NumExpr)(1)))) );
  CHECK( parse_str_error("_fun x x + 1")
         == "expected `(` after `_fun`" );
  CHECK( parse_str_error("_fun (1) x + 1")
         == "expected a name after `(` for `_fun`" );
  CHECK( parse_str_error("_fun (x 1) x + 1")
         == "expected `)` after argument for `_fun`" );
  
  CHECK( parse_str("f(1)")->equals(NEW(CallExpr)(NEW(VarExpr)("f"), NEW(NumExpr)(1))) );
  CHECK( parse_str("f(1)(3)*2")
         ->equals(NEW(MultExpr)(NEW(CallExpr)(NEW(CallExpr)(NEW(VarExpr)("f"),
                                                            NEW(NumExpr)(1)),
                                              NEW(NumExpr)(3)),
                                NEW(NumExpr)(2))) );
  CHECK( parse_str("fib(fib)(x+-1)")
         ->equals(NEW(CallExpr)(NEW(CallExpr)(NEW(VarExpr)("fib"),
                                              NEW(VarExpr)("fib")),
                                NEW(AddExpr)(NEW(VarExpr)("x"),
                                             NEW(NumExpr)(-1)))) );
  
  CHECK( parse_str_error("f(1 2)")
         == "expected a close parenthesis for a function call" );
}
