#include "value.hpp"

#include "env.hpp"
#include "expr.hpp"
#include <stdexcept>
#include "catch.hpp"

NumVal::NumVal(int rep) {
  this->rep = rep;
}

bool NumVal::equals(PTR(Val) other_val) {
  PTR(NumVal) other_num_val = CAST(NumVal)(other_val);
  if (other_num_val == nullptr)
    return false;
  else
    return rep == other_num_val->rep;
}

PTR(Val) NumVal::add_to(PTR(Val) other_val) {
  PTR(NumVal) other_num_val = CAST(NumVal)(other_val);
  if (other_num_val == nullptr)
    throw std::runtime_error("not a number");
  else
    return NEW(NumVal)(rep + other_num_val->rep);
}

PTR(Val) NumVal::mult_with(PTR(Val) other_val) {
  PTR(NumVal) other_num_val = CAST(NumVal)(other_val);
  if (other_num_val == nullptr)
    throw std::runtime_error("not a number");
  else
    return NEW(NumVal)(rep * other_num_val->rep);
}

bool NumVal::is_true() {
  throw std::runtime_error("not a boolean");
}

PTR(Val) NumVal::call(PTR(Val) actual_arg) {
  throw std::runtime_error("not a function");
}

PTR(Expr) NumVal::to_expr() {
  return NEW(NumExpr)(rep);
}

std::string NumVal::to_string() {
  return std::to_string(rep);
}

BoolVal::BoolVal(bool rep) {
  this->rep = rep;
}

bool BoolVal::equals(PTR(Val) other_val) {
  PTR(BoolVal) other_bool_val = CAST(BoolVal)(other_val);
  if (other_bool_val == nullptr)
    return false;
  else
    return rep == other_bool_val->rep;
}

PTR(Val) BoolVal::add_to(PTR(Val) other_val) {
  throw std::runtime_error("no adding booleans");
}

PTR(Val) BoolVal::mult_with(PTR(Val) other_val) {
  throw std::runtime_error("no multiplying booleans");
}

bool BoolVal::is_true() {
  return rep;
}

PTR(Val) BoolVal::call(PTR(Val) actual_arg) {
  throw std::runtime_error("not a function");
}

PTR(Expr) BoolVal::to_expr() {
  return NEW(BoolExpr)(rep);
}

std::string BoolVal::to_string() {
  if (rep)
    return "_true";
  else
    return "_false";
}

FunVal::FunVal(std::string formal_arg, PTR(Expr) body, PTR(Env) env) {
  this->formal_arg = formal_arg;
  this->body = body;
  this->env = env;
}

bool FunVal::equals(PTR(Val) other_val) {
  PTR(FunVal) other_fun_val = CAST(FunVal)(other_val);
  if (other_fun_val == nullptr)
    return false;
  else
    return ((formal_arg == other_fun_val->formal_arg)
            && body->equals(other_fun_val->body)
            && env->equals(other_fun_val->env));
}

PTR(Val) FunVal::add_to(PTR(Val) other_val) {
  throw std::runtime_error("no adding functions");
}

PTR(Val) FunVal::mult_with(PTR(Val) other_val) {
  throw std::runtime_error("no multiplying functions");
}

bool FunVal::is_true() {
  throw std::runtime_error("not a boolean");
}

PTR(Val) FunVal::call(PTR(Val) actual_arg) {
  return body->interp(NEW(ExtendedEnv)(formal_arg, actual_arg, env));
}

PTR(Expr) FunVal::to_expr() {
  return NEW(FunExpr)(formal_arg, body);
}

std::string FunVal::to_string() {
  return "[function]";
}

TEST_CASE( "values equals" ) {
  CHECK( (NEW(NumVal)(5))->equals(NEW(NumVal)(5)) );
  CHECK( ! (NEW(NumVal)(7))->equals(NEW(NumVal)(5)) );

  CHECK( (NEW(BoolVal)(true))->equals(NEW(BoolVal)(true)) );
  CHECK( ! (NEW(BoolVal)(true))->equals(NEW(BoolVal)(false)) );
  CHECK( ! (NEW(BoolVal)(false))->equals(NEW(BoolVal)(true)) );

  CHECK( ! (NEW(NumVal)(7))->equals(NEW(BoolVal)(false)) );
  CHECK( ! (NEW(BoolVal)(false))->equals(NEW(NumVal)(8)) );
  
  CHECK( (NEW(FunVal)("x", NEW(NumExpr)(8), Env::empty))
         ->equals(NEW(FunVal)("x", NEW(NumExpr)(8), Env::empty)) );
  CHECK( ! (NEW(FunVal)("x", NEW(NumExpr)(8), Env::empty))
         ->equals(NEW(FunVal)("y", NEW(NumExpr)(8), Env::empty)) );
  CHECK( ! (NEW(FunVal)("x", NEW(NumExpr)(8), Env::empty))
         ->equals(NEW(FunVal)("x", NEW(NumExpr)(10), Env::empty)) );
  CHECK( ! (NEW(FunVal)("x", NEW(NumExpr)(8), Env::empty))
         ->equals(NEW(NumVal)(8)) );

}

TEST_CASE( "add_to" ) {
  CHECK ( (NEW(NumVal)(5))->add_to(NEW(NumVal)(8))->equals(NEW(NumVal)(13)) );

  CHECK_THROWS_WITH ( (NEW(NumVal)(5))->add_to(NEW(BoolVal)(false)), "not a number" );
  CHECK_THROWS_WITH ( (NEW(BoolVal)(false))->add_to(NEW(NumVal)(0)),
                      "no adding booleans" );
  CHECK_THROWS_WITH ( (NEW(FunVal)("x", NEW(NumExpr)(8), Env::empty))->add_to(NEW(NumVal)(0)),
                      "no adding functions" );
}

TEST_CASE( "mult_with" ) {
  CHECK ( (NEW(NumVal)(5))->mult_with(NEW(NumVal)(8))->equals(NEW(NumVal)(40)) );

  CHECK_THROWS_WITH ( (NEW(NumVal)(5))->mult_with(NEW(BoolVal)(false)), "not a number" );
  CHECK_THROWS_WITH ( (NEW(BoolVal)(false))->mult_with(NEW(NumVal)(0)),
                      "no multiplying booleans" );
  CHECK_THROWS_WITH ( (NEW(FunVal)("x", NEW(NumExpr)(8), Env::empty))->mult_with(NEW(NumVal)(0)),
                      "no multiplying functions" );
}

TEST_CASE( "is_true" ) {
  CHECK ( (NEW(BoolVal)(true))->is_true() );
  CHECK ( ! (NEW(BoolVal)(false))->is_true() );

  CHECK_THROWS_WITH ( (NEW(NumVal)(5))->is_true(), "not a boolean" );
  CHECK_THROWS_WITH ( (NEW(FunVal)("x", NEW(NumExpr)(8), Env::empty))->is_true(), "not a boolean" );
}

TEST_CASE( "call" ) {
  CHECK( (NEW(FunVal)("x", NEW(NumExpr)(8), Env::empty))->call(NEW(NumVal)(0))
         ->equals(NEW(NumVal)(8)) );
  CHECK( (NEW(FunVal)("x", NEW(AddExpr)(NEW(VarExpr)("x"), NEW(NumExpr)(8)), Env::empty))
         ->call(NEW(NumVal)(1))
         ->equals(NEW(NumVal)(9)) );
  CHECK( (NEW(FunVal)("x",
                      NEW(AddExpr)(NEW(VarExpr)("x"), NEW(VarExpr)("z")),
                      NEW(ExtendedEnv)("z", NEW(NumVal)(8), Env::empty)))
         ->call(NEW(NumVal)(1))
         ->equals(NEW(NumVal)(9)) );

  CHECK_THROWS_WITH ( (NEW(NumVal)(5))->call(NEW(NumVal)(0)), "not a function" );
  CHECK_THROWS_WITH ( (NEW(BoolVal)(false))->call(NEW(NumVal)(0)), "not a function" );
}

TEST_CASE( "value to_expr" ) {
  CHECK( (NEW(NumVal)(5))->to_expr()->equals(NEW(NumExpr)(5)) );
  CHECK( (NEW(BoolVal)(true))->to_expr()->equals(NEW(BoolExpr)(true)) );
  CHECK( (NEW(BoolVal)(false))->to_expr()->equals(NEW(BoolExpr)(false)) );
  CHECK ( (NEW(FunVal)("x", NEW(NumExpr)(8), Env::empty))->to_expr()
          ->equals(NEW(FunExpr)("x", NEW(NumExpr)(8))) );
}

TEST_CASE( "value to_string" ) {
  CHECK( (NEW(NumVal)(5))->to_string() == "5" );
  CHECK( (NEW(BoolVal)(true))->to_string() == "_true" );
  CHECK( (NEW(BoolVal)(false))->to_string() == "_false" );
  CHECK ( (NEW(FunVal)("x", NEW(NumExpr)(8), Env::empty))->to_string() == "[function]" );
}

